import 'dart:async';
import 'dart:convert';

import 'package:verival/update.dart';
import 'package:verival/verival.dart' as verival;
import 'package:webinbox/webinbox.dart' as inbox;
import 'package:webroker/send.dart' as webroker;
import 'package:webroker/target.dart';
import 'package:yona_crypto/b64.dart';
import 'package:yona_crypto/crypto.dart';
import 'package:yona_messenger/peer.dart';

import 'addr.dart';
import 'msg.dart';

abstract class EventHandler {
  Future<void> onMsg(Msg msg);
}

class Node {
  final EventHandler _eh;
  final B64 sk;
  final Addr webinbox;
  final Peer peer;
  final Map<String, Completer<void>> _awaiters = {};

  Future<void> _onMsg(Msg msg) async {
    await _eh.onMsg(msg);
  }

  Future<Addr> lookup(Peer peer) async {
    var update =
        await verival.get(peer.verival.host, peer.verival.port, peer.pk);
    var webroker = Target.fromString(update.data["webroker"]!);
    var webinbox = Addr.fromString(update.data["webinbox"]!);
    peer.webroker = webroker;
    return webinbox;
  }

  Future<void> publish() async {
    var update = Update.create(peer.pk, DateTime.now().toUtc(), {
      "webroker": peer.webroker!.toString(),
      "webinbox": webinbox.toString(),
    });
    await verival.put(
        peer.verival.host, peer.verival.port, await update.encode(sk));
  }

  Node(this.sk, this.peer, this._eh, this.webinbox) {
    if (!(peer.pk == derivePK(sk))) {
      throw ArgumentError("keys are not related");
    }
  }

  Future<void> input(List<int> data) async {
    var input = Uri.splitQueryString(utf8.decode(data));
    var from = Peer.fromString(input["from"]!);
    if (from.webroker == null) {
      return;
    }
    var to = Peer.fromString(input["to"]!);
    if (from == to) {
      throw Exception("to is equal to from");
    }
    if (!(to == peer)) {
      return;
    }
    var key = await deriveKey(sk, from.pk);
    var payload =
        Uri.splitQueryString(utf8.decode(await decrypt(key, input["cipher"]!)));
    switch (payload["kind"]) {
      case "data":
        var msg = Msg(
          from,
          payload["data"]!,
          MsgID.fromString(payload["id"]!),
        );
        if (payload["ack"]?.toLowerCase() == "true") {
          var ack = Uri(queryParameters: {
            "kind": "ack",
            "msg": payload["id"]!,
          }).query;
          var key = await deriveKey(sk, from.pk);
          var cipher = await encrypt(key, utf8.encode(ack));
          var output = Uri(queryParameters: {
            "from": peer.toString(),
            "to": from.toString(),
            "cipher": cipher,
          }).query;
          await webroker.send(from.webroker!, utf8.encoder.convert(output));
        }
        await _onMsg(msg);
        break;
      case "ack":
        var msgId = MsgID.fromString(payload["msg"]!);
        _awaiters[msgId.toString()]!.complete();
        break;
      default:
        throw UnsupportedError("unsupported data kind");
    }
  }

  Future<bool> send(Peer to, String data,
      [bool ack = true, MsgID? msgId]) async {
    if (peer.webroker == null) {
      throw Exception("no webroker");
    }
    msgId ??= MsgID.now();
    Timer? timer;
    try {
      var webinboxAddr = await lookup(to);
      if (to.webroker == null) ack = false;
      var payload = Uri(queryParameters: {
        "kind": "data",
        "ack": ack.toString(),
        "data": data,
        "id": msgId.toString()
      }).query;
      var key = await deriveKey(sk, to.pk);
      var cipher = await encrypt(key, utf8.encode(payload));

      if (to.webroker == null) {
        var msg = inbox.Msg(peer.pk, to.pk, cipher);
        var sig = sign(sk, utf8.encoder.convert(msg.data));
        await inbox.put(
            Uri(
                scheme: "http",
                host: webinboxAddr.host,
                port: webinboxAddr.port),
            msg,
            sig);
        return false;
      } else {
        var output = Uri(queryParameters: {
          "from": peer.toString(),
          "to": to.toString(),
          "cipher": cipher,
        }).query;
        await webroker.send(to.webroker!, utf8.encoder.convert(output));
        if (ack) {
          var completer = Completer<void>();
          timer = Timer(Duration(seconds: 3), () {
            completer.completeError(TimeoutException("ack timed out"));
          });
          _awaiters[msgId.toString()] = completer;
          await completer.future;
        }
        return true;
      }
    } finally {
      _awaiters.remove(msgId.toString());
      timer?.cancel();
    }
  }
}
